/*******************************************
 * Popup dialog handling routines start here
 *
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <stdio.h>
#include "string.h"

#include "util.h"
#include "resource.h"

#define EF_MAX_VISIBLE 50
#define EF_MAX 256
#define BUTTON_LABEL "OK"
#define GUTTER_HT 5
#define GUTTER_WID 5

void make_dialog(void);
char *dialog_event_loop(char *label);

extern Display *dpy;
extern int scrn;
extern Window root;
extern Window mainWin;
extern GC theGc;
extern XFontStruct *theFont;
extern Colormap cmap;
extern unsigned long pixels[];		/* 128  SAM colours */


static Window dlg_win, entry_win, ok_win;
static int dlg_wid, dlg_ht, dlg_x, dlg_y;
static int char_width, font_ht;
static int label_wid, label_ht, entry_wid, entry_ht, ok_wid, ok_ht;
static int cursor_pos = 0, cur_win_pos = 0;
static char dlg_string[EF_MAX];

/*******************************************
 * Popup dialog handling routines start here
 *
 */

/* Create the popup dialog box */
void
make_dialog(void)
{
	XWMHints xwmh;
	XSizeHints xsh;
	XSetWindowAttributes xswa;
	XTextProperty dlgName;
	char *dlg_name;

	/* first work out the necessary geometries */
	font_ht = theFont->ascent + theFont->descent;
	/* only works with fixed width fonts! */
	char_width = XTextWidth(theFont, " ", 1);

	label_wid = char_width * EF_MAX_VISIBLE;
	label_ht = font_ht;
	entry_wid = (char_width * EF_MAX_VISIBLE) + 4;
	entry_ht = font_ht + 4;
	ok_wid = (char_width * strlen(BUTTON_LABEL)) + 4;
	ok_ht = font_ht + 4;
	dlg_wid = entry_wid + (2 * GUTTER_WID);
	dlg_ht = label_ht + entry_ht + ok_ht + (4 * GUTTER_HT);

	/* window creation */
	dlg_win = XCreateSimpleWindow(dpy, root, dlg_x, dlg_y, dlg_wid, dlg_ht, 
		0, pixels[0], pixels[112]);
	entry_win = XCreateSimpleWindow(dpy, dlg_win, GUTTER_WID, 
		(2 * GUTTER_HT) + label_ht, entry_wid, entry_ht, 0,
		pixels[0], pixels[112]);
	ok_win = XCreateSimpleWindow(dpy, dlg_win, GUTTER_WID, 
		(3 * GUTTER_HT) + label_ht + entry_ht, ok_wid, ok_ht, 0,
		pixels[0], pixels[112]);


	/* tell the window manager */
	xsh.flags = PSize | PMinSize | PMaxSize | PPosition;
	xsh.min_width  = dlg_wid; xsh.min_height = dlg_ht;
	xsh.max_width  = dlg_wid; xsh.max_height = dlg_ht;

	xwmh.initial_state = NormalState;
	xwmh.input = True;
	xwmh.flags = StateHint | InputHint;

	dlg_name = "simcoupe popup";
	if (XStringListToTextProperty(&dlg_name, 1, &dlgName) == 0) {
		simcoupe_mesg(SIMCOUPE_ERR, "can't create a TextProperty!");
		return;
	}

	xswa.save_under = True;

	if (GetCfg(priv_cmap))
		XSetWindowColormap(dpy, dlg_win, cmap);
	
	XSelectInput(dpy, dlg_win, ExposureMask | KeyPressMask | EnterWindowMask);
	XSelectInput(dpy, ok_win, ButtonPressMask | ButtonReleaseMask);

	XChangeWindowAttributes(dpy, dlg_win, CWSaveUnder, &xswa);
#if 0
	XSetTransientForHint(dpy, dlg_win, borderWin);
#endif
	XSetWMProperties(dpy, dlg_win, &dlgName, NULL,
		&dlg_name, 1, &xsh, &xwmh, NULL);

	/* subwindows must be mapped & raised */
	XMapRaised(dpy, entry_win);
	XMapRaised(dpy, ok_win);
}


static void
paint_entry_field()
{
	int real_curpos;

	/* draw the string */
	XSetForeground(dpy, theGc, pixels[0]);
	XSetBackground(dpy, theGc, pixels[112]);
	XDrawImageString(dpy, entry_win, theGc, 2, 2 + theFont->ascent, 
		dlg_string + cur_win_pos, strlen(dlg_string + cur_win_pos));

	/* draw the cursor */
	real_curpos = (cursor_pos - cur_win_pos) * char_width + 2;
	XFillRectangle(dpy, entry_win, theGc,
		real_curpos, 2, char_width, entry_ht - 4);

	/* blank the rest of the window */
	XClearArea(dpy, entry_win, real_curpos + char_width, 2,
		(EF_MAX_VISIBLE * char_width) - (real_curpos + char_width + 2),
		font_ht, False);
}


static void
paint_dialog(char *label)
{
	/* draw entry field and button borders */
	XSetForeground(dpy, theGc, pixels[15]);
	XDrawLine(dpy, entry_win, theGc, 0, entry_ht - 1, 0, 0);
	XDrawLine(dpy, entry_win, theGc, 0, 0, entry_wid - 1, 0);
	XSetForeground(dpy, theGc, pixels[119]);
	XDrawLine(dpy, entry_win, theGc, entry_wid - 1, 0,
		entry_wid - 1, entry_ht - 1);
	XDrawLine(dpy, entry_win, theGc, entry_wid - 1, entry_ht - 1,
		0, entry_ht - 1);

	XSetForeground(dpy, theGc, pixels[119]);
	XDrawLine(dpy, ok_win, theGc, 0, ok_ht - 1, 0, 0);
	XDrawLine(dpy, ok_win, theGc, 0, 0, ok_wid - 1, 0);
	XSetForeground(dpy, theGc, pixels[15]);
	XDrawLine(dpy, ok_win, theGc, ok_wid - 1, 0, ok_wid - 1, ok_ht - 1);
	XDrawLine(dpy, ok_win, theGc, ok_wid - 1, ok_ht - 1, 0, ok_ht - 1);

	/* paint the label & the button label */
	XSetForeground(dpy, theGc, pixels[0]);
	XDrawString(dpy, dlg_win, theGc, GUTTER_WID, GUTTER_HT + theFont->ascent,
		label, strlen(label));
	XDrawString(dpy, ok_win, theGc, 2, 2 + theFont->ascent,
		BUTTON_LABEL, strlen(BUTTON_LABEL));

	paint_entry_field();
}


static void
shadow_in(Window w, int wid, int ht)
{
	XSetForeground(dpy, theGc, pixels[15]);
	XDrawLine(dpy, w, theGc, 0, ht - 1, 0, 0);
	XDrawLine(dpy, w, theGc, 0, 0, wid - 1, 0);
	XSetForeground(dpy, theGc, pixels[119]);
	XDrawLine(dpy, w, theGc, wid - 1, 0, wid - 1, ht - 1);
	XDrawLine(dpy, w, theGc, wid - 1, ht - 1, 0, ht - 1);
}

static void
shadow_out(Window w, int wid, int ht)
{
	XSetForeground(dpy, theGc, pixels[119]);
	XDrawLine(dpy, w, theGc, 0, ht - 1, 0, 0);
	XDrawLine(dpy, w, theGc, 0, 0, wid - 1, 0);
	XSetForeground(dpy, theGc, pixels[15]);
	XDrawLine(dpy, w, theGc, wid - 1, 0, wid - 1, ht - 1);
	XDrawLine(dpy, w, theGc, wid - 1, ht - 1, 0, ht - 1);
}


/* Process a keypress in the dialog entry field.  Return a value of 1 if
 * we're finished typing text, 0 otherwise.
 */
static int
process_ef_key(XKeyEvent *kev)
{
	char buf[3];
	KeySym ks;

	XLookupString(kev, buf, 2, &ks, NULL);

	switch (ks) {
		case XK_Return:
			return 1;
			break;
		case XK_Escape:
			dlg_string[0] = '\0';
			return 1;
			break;
		case XK_BackSpace: case XK_Delete:
			if (cursor_pos > 0) cursor_pos--;
			dlg_string[cursor_pos] = '\0';
			if (cursor_pos < cur_win_pos) cur_win_pos--;
			break;
		default:
			if (ks < 128 && ks > 31) {
				/* add a character */
				dlg_string[cursor_pos] = (char) ks;
				if (cursor_pos < EF_MAX - 1) cursor_pos++;
				if (cursor_pos >= cur_win_pos + EF_MAX_VISIBLE) cur_win_pos++;
				dlg_string[cursor_pos] = '\0';
			}
			break;
	}

	paint_entry_field();
	XFlush(dpy);

	return 0;
}


/* Pop up the dialog box, and process events until Return is pressed or
 * the OK button is clicked.  Then pop down the dialog box and return the
 * string in the entry field.
 */
char *
dialog_event_loop(char *label)
{
	int finished = 0;
	XEvent xev;
	Window r_ret, w_ret;
	int dx, dy, winx, winy, mask_ret; 

	/* reset everything */
	dlg_string[0] = '\0'; cursor_pos = 0; cur_win_pos = 0;

	/* erase the old label */
	XClearArea(dpy, dlg_win, GUTTER_WID, GUTTER_HT, label_wid, label_ht, False);

	/* move the dialog window to under the pointer */
	if (!XQueryPointer(dpy, root, &r_ret, &w_ret, &dx, &dy,
		&winx, &winy, &mask_ret)) {
		dx = 0; dy = 0;
	} else {
		dx -= dlg_wid / 2; dy -= dlg_ht / 2;
	}
	if (dx < 0) dx = 0;
	if (dy < 0) dy = 0;
	XMoveWindow(dpy, dlg_win, dx, dy);

	XMapRaised(dpy, dlg_win);

	while (!finished) {
		XNextEvent(dpy, &xev);
		switch(xev.type) {
			case Expose:
				paint_dialog(label);
				break;
			case ButtonPress:
				if (xev.xbutton.window == ok_win) {
					shadow_in(xev.xbutton.window, ok_wid, ok_ht);
				}
				break;
			case ButtonRelease:
				if (xev.xbutton.window == ok_win) {
					shadow_out(xev.xbutton.window, ok_wid, ok_ht);
					finished = 1;
				}
				break;
			case KeyPress:
				if (xev.xkey.window == entry_win ||
					xev.xkey.window == dlg_win) {
					finished = process_ef_key((XKeyEvent *)&xev);
				}
				break;
			case EnterNotify:
				if (xev.xcrossing.window == dlg_win)
					XAutoRepeatOn(dpy);
				break;
			default:	/* ignore any other events */
				break;
					
		}
	}

	XUnmapWindow(dpy, dlg_win);
	XFlush(dpy);

	return dlg_string;
}
